@@ -7,31 +7,16 @@ import bdb
 import gc
 import os
 import sys
+from functools import partial
 from types import TracebackType
 
+from pudb.lowlevel import decode_lines
 from pudb.settings import load_config, save_config
+from pudb.py3compat import PY3, raw_input, execfile
+
 CONFIG = load_config()
 save_config(CONFIG)
 
-from pudb.py3compat import PY3, raw_input
-if PY3:
-    _next = "__next__"
-else:
-    _next = "next"
-
-try:
-    from functools import partial
-except ImportError:
-    def partial(func, *args, **keywords):
-        def newfunc(*fargs, **fkeywords):
-            newkeywords = keywords.copy()
-            newkeywords.update(fkeywords)
-            return func(*(args + fargs), **newkeywords)
-        newfunc.func = func
-        newfunc.args = args
-        newfunc.keywords = keywords
-        return newfunc
-
 HELP_TEXT = """\
 Welcome to PuDB, the Python Urwid debugger.
 -------------------------------------------
@@ -77,12 +62,12 @@ Keys:
 
     Ctrl-l - redraw screen
 
-Command line-related:
-    ! - invoke configured python command line in current environment
-    Ctrl-x - toggle inline command line focus
+Shell-related:
+    ! - open the external shell (configured in the settings)
+    Ctrl-x - toggle the internal shell focus
 
-    +/- - grow/shrink inline command line (active in command line history)
-    _/= - minimize/maximize inline command line (active in command line history)
+    +/- - grow/shrink inline shell (active in command line history)
+    _/= - minimize/maximize inline shell (active in command line history)
 
     Ctrl-v - insert newline
     Ctrl-n/p - browse command line history
@@ -555,10 +540,7 @@ class FileSourceCodeProvider(SourceCodeProvider):
         self.file_name = debugger.canonic(file_name)
 
     def __eq__(self, other):
-        return (
-                type(self) == type(other)
-                and
-                self.file_name == other.file_name)
+        return type(self) == type(other) and self.file_name == other.file_name
 
     def identifier(self):
         return self.file_name
@@ -585,18 +567,8 @@ class FileSourceCodeProvider(SourceCodeProvider):
         try:
             from linecache import getlines
             lines = getlines(self.file_name)
-
-            from pudb.lowlevel import detect_encoding
-            source_enc, _ = detect_encoding(getattr(iter(lines), _next))
-
-            decoded_lines = []
-            for l in lines:
-                if hasattr(l, "decode"):
-                    decoded_lines.append(l.decode(source_enc))
-                else:
-                    decoded_lines.append(l)
-
-            return format_source(debugger_ui, decoded_lines, set(breakpoints))
+            return format_source(
+                    debugger_ui, list(decode_lines(lines)), set(breakpoints))
         except:
             from pudb.lowlevel import format_exception
             debugger_ui.message("Could not load source file '%s':\n\n%s" % (
@@ -631,22 +603,8 @@ class DirectSourceCodeProvider(SourceCodeProvider):
     def get_lines(self, debugger_ui):
         from pudb.source_view import format_source
 
-        lines = self.code.split("\n")
-
-        from pudb.lowlevel import detect_encoding
-        source_enc, _ = detect_encoding(getattr(iter(lines), _next))
-
-        decoded_lines = []
-        for i, l in enumerate(lines):
-            if hasattr(l, "decode"):
-                l = l.decode(source_enc)
-
-            if i+1 < len(lines):
-                l += "\n"
-
-            decoded_lines.append(l)
-
-        return format_source(debugger_ui, decoded_lines, set())
+        lines = self.code.splitlines(True)
+        return format_source(debugger_ui, list(decode_lines(lines)), set())
 
 # }}}
 
@@ -1799,12 +1757,6 @@ class DebuggerUI(FrameVarInfoKeeper):
         def run_external_cmdline(w, size, key):
             self.screen.stop()
 
-            if not hasattr(self, "have_been_to_cmdline"):
-                self.have_been_to_cmdline = True
-                first_cmdline_run = True
-            else:
-                first_cmdline_run = False
-
             curframe = self.debugger.curframe
 
             import pudb.shell as shell
@@ -1814,11 +1766,30 @@ class DebuggerUI(FrameVarInfoKeeper):
                 runner = shell.run_bpython_shell
             elif CONFIG["shell"] == "ptpython" and shell.HAVE_PTPYTHON:
                 runner = shell.run_ptpython_shell
-            else:
+            elif CONFIG["shell"] == "classic":
                 runner = shell.run_classic_shell
+            else:
+                try:
+                    if not shell.custom_shell_dict:  # Only execfile once
+                        from os.path import expanduser
+                        execfile(
+                                expanduser(CONFIG["shell"]), shell.custom_shell_dict)
+                except:
+                    print("Error when importing custom shell:")
+                    from traceback import print_exc
+                    print_exc()
+                    print("Falling back to classic shell")
+                    runner = shell.run_classic_shell
+                else:
+                    if "pudb_shell" not in shell.custom_shell_dict:
+                        print("%s does not contain a function named pudb_shell at "
+                              "the module level." % CONFIG["shell"])
+                        print("Falling back to classic shell")
+                        runner = shell.run_classic_shell
+                    else:
+                        runner = shell.custom_shell_dict['pudb_shell']
 
-            runner(curframe.f_locals, curframe.f_globals,
-                    first_cmdline_run)
+            runner(curframe.f_globals, curframe.f_locals)
 
             self.screen.start()
 
@@ -2129,7 +2100,7 @@ class DebuggerUI(FrameVarInfoKeeper):
                 self.message("Package 'pygments' not found. "
                         "Syntax highlighting disabled.")
 
-        WELCOME_LEVEL = "e031"  # noqa
+        WELCOME_LEVEL = "e032"  # noqa
         if CONFIG["seen_welcome"] < WELCOME_LEVEL:
             CONFIG["seen_welcome"] = WELCOME_LEVEL
             from pudb import VERSION
@@ -2146,6 +2117,13 @@ class DebuggerUI(FrameVarInfoKeeper):
                     "(invoked by hitting '?' after this message) should get you "
                     "on your way.\n"
 
+                    "\nChanges in version 2017.1.3:\n\n"
+                    "- Add handling of safely_stringify_for_pudb to allow custom \n"
+                    "  per-type stringification.\n"
+                    "- Add support for custom shells.\n"
+                    "- Better support for 2-wide characters in the var view.\n"
+                    "- Bug fixes.\n"
+
                     "\nChanges in version 2017.1.2:\n\n"
                     "- Bug fixes.\n"
 
